use std::cell::RefCell;
use std::rc::Rc;

pub trait CommandId {
    fn command_id(&self) -> u64;
}

impl CommandId for u64 {
    fn command_id(&self) -> u64 {
        *self
    }
}

pub trait WithCommandData {
    fn command_data(&self) -> &CommandData;
}

pub trait CommandInfo: CommandId {
    fn name(&self) -> &str;
}

pub struct CommandData {
    command_id: u64,
    name: String,
}

impl CommandData {
    pub fn new<TName: Into<String>>(command_id: u64, name: TName) -> Self {
        CommandData {
            command_id,
            name: name.into(),
        }
    }
}

impl CommandId for CommandData {
    fn command_id(&self) -> u64 {
        self.command_id
    }
}

impl CommandInfo for CommandData {
    fn name(&self) -> &str {
        &self.name
    }
}

impl<T: WithCommandData> CommandId for T {
    fn command_id(&self) -> u64 {
        self.command_data().command_id
    }
}

impl<T: WithCommandData> CommandInfo for T {
    fn name(&self) -> &str {
        &self.command_data().name
    }
}

#[derive(Debug, PartialEq, Clone, Copy)]
pub enum CommandResult {
    Continue,
    Done,
    Cancel,
}

pub trait Command: WithCommandData {
    fn init(&mut self);
    fn step(&mut self) -> CommandResult;
    fn finish(&mut self, result: CommandResult);
}

pub struct GenericCommand<TConfig, TState> {
    command_data: CommandData,
    pub config: TConfig,
    pub state: TState,
}

impl<TConfig, TState> GenericCommand<TConfig, TState>
where
    TConfig: 'static,
    GenericCommand<TConfig, TState>: Command,
    TState: Default + 'static,
{
    pub fn new(command_data: CommandData, config: TConfig) -> Self {
        Self {
            command_data,
            config,
            state: TState::default(),
        }
    }

    pub fn new_ref(command_data: CommandData, config: TConfig) -> CommandRef {
        Rc::new(RefCell::new(Self::new(command_data, config)))
    }
}

impl<TConfig, TState> WithCommandData for GenericCommand<TConfig, TState> {
    fn command_data(&self) -> &CommandData {
        &self.command_data
    }
}

macro_rules! command_impl {
    (
        $command:ident<$tconfig:ty,$tstate:ty>;
        $init_self:ident => $init:block;
        $step_self:ident => $step:block;
        $finish_self:ident, $finish_result:ident => $finish:block
    ) => {
        pub type $command = GenericCommand<$tconfig, $tstate>;

        impl $crate::command::Command for $command {
            fn init(&mut $init_self) {
                use $crate::command::CommandInfo;
                debug!(
                    "{}({}) init",
                    stringify!($command_st),
                    $init_self.name()
                );
                $init;
            }

            fn step(&mut $step_self) -> $crate::command::CommandResult {
                use $crate::command::CommandInfo;
                debug!(
                    "{}({}) step",
                    stringify!($command_st),
                    $step_self.name()
                );
                $step
            }

            fn finish(&mut $finish_self, $finish_result: $crate::command::CommandResult) {
                use $crate::command::CommandInfo;
                debug!(
                    "{}({}) finish({:?}",
                    stringify!($command_st),
                    $finish_self.name(),
                    $finish_result,
                );
                $finish;
            }

        }
    };
    (
        $command:ident<$tconfig:ty,$tstate:ty>;
        $init_self:ident => $init:block;
        $step_self:ident => $step:block;
        $finish_self:ident => $finish:block
    ) => {
        command_impl! { $command<$tconfig,$tstate>;
        $init_self => $init; $step_self => $step; $finish_self, res => {} }
    };
    (
        $command:ident<$tconfig:ty>;
        $init_self:ident => $init:block
    ) => {
        command_impl! { $command<$tconfig,()>;
        $init_self => $init; self => { $crate::command::CommandResult::Done }; self => {} }
    };
}

pub type CommandRef = Rc<RefCell<dyn Command>>;

pub struct CommandGroup {
    commands: Vec<CommandRef>,
    command_data: CommandData,
    current_idx: Option<usize>,
    has_initialized: bool,
    has_cancelled: bool,
}

impl CommandGroup {
    pub fn new<TCommands>(data: CommandData, commands: TCommands) -> Self
    where
        TCommands: IntoIterator<Item = CommandRef>,
    {
        CommandGroup {
            commands: commands.into_iter().collect(),
            command_data: data,
            current_idx: None,
            has_initialized: false,
            has_cancelled: false,
        }
    }
}

impl WithCommandData for CommandGroup {
    fn command_data(&self) -> &CommandData {
        &self.command_data
    }
}

impl Command for CommandGroup {
    fn init(&mut self) {
        if self.current_idx.is_some() {
            error!("CommandGroup init called twice");
        }
        self.current_idx = Some(0);
        self.has_initialized = false;
        self.has_cancelled = false;
        debug!("CommandGroup {} init", self.command_data().name());
    }

    fn step(&mut self) -> CommandResult {
        use self::CommandResult::*;
        if self.has_cancelled {
            debug!("CommandGroup step called after cancelled");
            return CommandResult::Cancel;
        }
        let mut current_idx = if let Some(c) = self.current_idx {
            c
        } else {
            error!("CommandGroup step called before init");
            self.has_cancelled = true;
            return CommandResult::Cancel;
        };
        let mut result: CommandResult;
        loop {
            if current_idx >= self.commands.len() {
                debug!("CommandGroup out of commands");
                return CommandResult::Done;
            }
            let mut command = self.commands[current_idx].borrow_mut();
            if !self.has_initialized {
                debug!(
                    "CommandGroup init {0} ({1})",
                    current_idx,
                    command.command_data().name()
                );
                command.init();
                self.has_initialized = true;
            }
            result = command.step();
            if result != Continue {
                debug!(
                    "CommandGroup {2:?} {0} ({1}) ",
                    current_idx,
                    command.command_data().name(),
                    result
                );
                command.finish(result)
            }
            match result {
                Done => {
                    current_idx += 1;
                    self.has_initialized = false;
                }
                Cancel => {
                    self.has_cancelled = true;
                    break;
                }
                _ => break,
            }
        }
        self.current_idx = Some(current_idx);
        CommandResult::Continue
    }

    fn finish(&mut self, result: CommandResult) {
        if result == CommandResult::Cancel {
            if let Some(command) = self
                .current_idx
                .and_then(|current_idx| self.commands.get(current_idx))
            {
                command.borrow_mut().finish(result);
            }
        }
        self.current_idx = None;
    }
}

pub struct CommandRun {
    initialized: bool,
    cancelled: bool,
    command: CommandRef,
}

impl CommandId for CommandRun {
    fn command_id(&self) -> u64 {
        self.command.borrow().command_data().command_id()
    }
}

impl CommandRun {
    pub fn initialized(&self) -> bool {
        self.initialized
    }

    pub fn cancelled(&self) -> bool {
        self.cancelled
    }

    pub fn cancel(&mut self) -> bool {
        if self.cancelled {
            false
        } else {
            self.cancelled = true;
            true
        }
    }
}

pub struct CommandScheduler {
    running_commands: Vec<CommandRun>,
}

impl CommandScheduler {
    pub fn new() -> Self {
        CommandScheduler {
            running_commands: Vec::new(),
        }
    }

    pub fn get_command_run<T: CommandId>(&self, id: T) -> Option<&CommandRun> {
        let id = id.command_id();
        for command_run in &self.running_commands {
            if command_run.command_id() == id {
                return Some(command_run);
            }
        }
        None
    }

    fn get_command_run_mut<T: CommandId>(&mut self, id: T) -> Option<&mut CommandRun> {
        let id = id.command_id();
        for command_run in &mut self.running_commands {
            if command_run.command_id() == id {
                return Some(command_run);
            }
        }
        None
    }

    pub fn is_running<T: CommandId>(&self, id: T) -> bool {
        self.get_command_run(id).is_some()
    }

    /**
     * Starts running a command.
     * Returns true if the command was started, false if it was not started because it was already running
     */
    pub fn start(&mut self, command: CommandRef) -> bool {
        if self.is_running(command.borrow().command_data().command_id()) {
            false
        } else {
            self.running_commands.push(CommandRun {
                initialized: false,
                cancelled: false,
                command,
            });
            true
        }
    }

    pub fn cancel<T: CommandId>(&mut self, id: T) -> bool {
        self.get_command_run_mut(id).map_or(false, |command_run| {
            command_run.cancelled = true;
            true
        })
    }

    pub fn execute(&mut self) {
        self.running_commands.drain_filter(|command_run| {
            let mut command = command_run.command.borrow_mut();
            if !command_run.initialized && !command_run.cancelled {
                {
                    let command_data = command.command_data();
                    debug!(
                        "initialzing command {} (id {})",
                        command_data.name(),
                        command_data.command_id()
                    );
                }
                command.init();
                command_run.initialized = true;
            }
            let result = if command_run.initialized {
                command.step()
            } else {
                CommandResult::Cancel
            };
            if result != CommandResult::Continue {
                command.finish(result);
                true
            } else {
                false
            }
        });
    }
}
