use std::cell::RefCell;
use std::rc::Rc;
use std::thread;
use std::time::Duration;

use command::{CommandData, CommandGroup, CommandRef, CommandScheduler};
use drive::{Drive, DriveRef};

mod cmds {
    use command::{Command, CommandResult, GenericCommand};
    use drive::DriveRef;
    use self::CommandResult::*;

    pub struct DriveCommandConfig {
        pub drive: DriveRef,
    }

    pub struct DriveCommandState {
        iterations: u64,
    }

    impl Default for DriveCommandState {
        fn default() -> Self {
            DriveCommandState { iterations: 0 }
        }
    }

    pub type DriveCommand = GenericCommand<DriveCommandConfig, DriveCommandState>;

    impl Command for DriveCommand {
        fn init(&mut self) {
            self.state = DriveCommandState::default();
        }
        fn step(&mut self) -> CommandResult {
            self.config.drive.borrow_mut().drive_powers(1.0, 1.0);
            self.state.iterations += 1;
            if self.state.iterations >= 10 {
                Done
            } else {
                Continue
            }
        }
        fn finish(&mut self, res: CommandResult) {
            self.config.drive.borrow_mut().drive_powers(0.0, 0.0);
        }
    }

    command_impl! {
        LogCmd<()>; self => { info!("{}", "LogCmd") }
    }
}

pub struct Robot {
    drive: DriveRef,
    command_scheduler: CommandScheduler,
}

const ITER_PERIOD: Duration = Duration::from_nanos(1_000_000_000 / 50);

impl Robot {
    pub fn new() -> Self {
        Robot {
            drive: Drive::new_ref(),
            command_scheduler: CommandScheduler::new(),
        }
    }

    pub fn run(mut self) {
        warn!("Starting RobotRS");

        let command = cmds::DriveCommand::new_ref(
            CommandData::new(0, "DriveCommand"),
            cmds::DriveCommandConfig {
                drive: self.drive.clone(),
            },
        );
        let log_command = cmds::LogCmd::new_ref(CommandData::new(2, "LogCommand"), ());
        let command_group = CommandGroup::new(
            CommandData::new(1, "Group"),
            vec![command, log_command],
        );
        let command_ref: CommandRef = Rc::new(RefCell::new(command_group));

        self.command_scheduler.start(command_ref);

        loop {
            self.execute();
            thread::sleep(ITER_PERIOD);
        }
    }

    fn execute(&mut self) {
        self.command_scheduler.execute();
    }
}
